/* Name: asmcommon.inc
 * Project: V-USB, virtual USB port for Atmel's(r) AVR(r) microcontrollers
 * Author: Christian Starkjohann
 * Creation Date: 2007-11-05
 * Tabsize: 4
 * Copyright: (c) 2007 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: GNU GPL v2 (see License.txt), GNU GPL v3 or proprietary (CommercialLicense.txt)
 * Revision: $Id$
 */

/* Do not link this file! Link usbdrvasm.S instead, which includes the
 * appropriate implementation!
 */

/*
General Description:
This file contains assembler code which is shared among the USB driver
implementations for different CPU cocks. Since the code must be inserted
in the middle of the module, it's split out into this file and #included.

Jump destinations called from outside:
    sofError: Called when no start sequence was found.
    se0: Called when a package has been successfully received.
    overflow: Called when receive buffer overflows.
    doReturn: Called after sending data.

Outside jump destinations used by this module:
    waitForJ: Called to receive an already arriving packet.
    sendAckAndReti:
    sendNakAndReti:
    sendCntAndReti:
    usbSendAndReti:

The following macros must be defined before this file is included:
    .macro POP_STANDARD
    .endm
    .macro POP_RETI
    .endm
*/

#define token   x1

overflow:
    ldi     x2, 1<<USB_INTR_PENDING_BIT
    USB_STORE_PENDING(x2)       ; clear any pending interrupts
ignorePacket:
    clr     token
    rjmp    storeTokenAndReturn

;----------------------------------------------------------------------------
; Processing of received packet (numbers in brackets are cycles after center of SE0)
;----------------------------------------------------------------------------
;This is the only non-error exit point for the software receiver loop
;we don't check any CRCs here because there is no time left.
se0:
    subi    cnt, USB_BUFSIZE    ;[5]
    neg     cnt                 ;[6]
    sub     YL, cnt             ;[7]
    sbci    YH, 0               ;[8]
    ldi     x2, 1<<USB_INTR_PENDING_BIT ;[9]
    USB_STORE_PENDING(x2)       ;[10] clear pending intr and check flag later. SE0 should be over.
    ld      token, y            ;[11]
    cpi     token, USBPID_DATA0 ;[13]
    breq    handleData          ;[14]
    cpi     token, USBPID_DATA1 ;[15]
    breq    handleData          ;[16]
    lds     shift, usbDeviceAddr;[17]
    ldd     x2, y+1             ;[19] ADDR and 1 bit endpoint number
    lsl     x2                  ;[21] shift out 1 bit endpoint number
    cpse    x2, shift           ;[22]
    rjmp    ignorePacket        ;[23]
/* only compute endpoint number in x3 if required later */
#if USB_CFG_HAVE_INTRIN_ENDPOINT || USB_CFG_IMPLEMENT_FN_WRITEOUT
    ldd     x3, y+2             ;[24] endpoint number + crc
    rol     x3                  ;[26] shift in LSB of endpoint
#endif
    cpi     token, USBPID_IN    ;[27]
    breq    handleIn            ;[28]
    cpi     token, USBPID_SETUP ;[29]
    breq    handleSetupOrOut    ;[30]
    cpi     token, USBPID_OUT   ;[31]
    brne    ignorePacket        ;[32] must be ack, nak or whatever
;   rjmp    handleSetupOrOut    ; fallthrough


;Setup and Out are followed by a data packet two bit times (16 cycles) after
;the end of SE0. The sync code allows up to 40 cycles delay from the start of
;the sync pattern until the first bit is sampled. That's a total of 56 cycles.
handleSetupOrOut:               ;[32]
#if USB_CFG_IMPLEMENT_FN_WRITEOUT   /* if we have data for endpoint != 0, set usbCurrentTok to address */
    andi    x3, 0xf             ;[32]
    breq    storeTokenAndReturn ;[33]
    mov     token, x3           ;[34] indicate that this is endpoint x OUT
#endif
storeTokenAndReturn:
    sts     usbCurrentTok, token;[35]
doReturn:
    POP_STANDARD                ;[37] 12...16 cycles
    USB_LOAD_PENDING(YL)        ;[49]
    sbrc    YL, USB_INTR_PENDING_BIT;[50] check whether data is already arriving
    rjmp    waitForJ            ;[51] save the pops and pushes -- a new interrupt is already pending
sofError:
    POP_RETI                    ;macro call
    reti


handleIn:
    lds x1,CurWritePos
    ldi x3,USB_BUFSIZE+IN_SIZE
    mul x1,x3
    mov x1,r0 ;rgCurWriteAddrL=CurWritePos*(11+3)
    ldi x2,0
    
    subi x1,lo8(-(FIFOBuf))
    sbci x2,hi8(-(FIFOBuf))
    mov YL,x1
    mov YH,x2
    lds x3,usbRxBuf+0
    st  Y+,x3
    lds x3,usbRxBuf+1
    st  Y+,x3
    lds x3,usbRxBuf+2
    st  Y+,x3
    rjmp doReturn  //Устанавливаем адрес для приема данных DATAx

handleData:
    lds x1,CurWritePos
    ldi x3,USB_BUFSIZE+IN_SIZE
    mul x1,x3
    mov x1,r0 ;rgCurWriteAddrL=CurWritePos*(11+3)
    ldi x3,IN_SIZE
    add x1,x3 ;rgCurWriteAddrL=CurWritePos*(11+3)+3
    ldi x2,0

    subi x1,lo8(-(FIFOBuf))
    sbci x2,hi8(-(FIFOBuf))

    mov YL,x1
    mov YH,x2

    lds x3,usbRxBuf+0
    st  Y+,x3

    lds x3,usbRxBuf+1
    st  Y+,x3
    lds x3,usbRxBuf+2
    st  Y+,x3
    lds x3,usbRxBuf+3
    st  Y+,x3
    lds x3,usbRxBuf+4
    st  Y+,x3
    lds x3,usbRxBuf+5
    st  Y+,x3
    lds x3,usbRxBuf+6
    st  Y+,x3
    lds x3,usbRxBuf+7
    st  Y+,x3
    lds x3,usbRxBuf+8
    st  Y+,x3
    lds x3,usbRxBuf+9
    st  Y+,x3
    lds x3,usbRxBuf+10
    st  Y+,x3
    /*  */
    lds x1,CurWritePos   
    subi x1,lo8(-(1))      //—двигаем буфер на одну позицию
    sts CurWritePos,x1
    cpi x1,lo8(FIFO_BUF_SIZE)   //сброс счетчика FIFO буфера при достижении максимума
    brlo SetWritePos
    ldi x1,0
    sts CurWritePos,x1
SetWritePos:
    rjmp  doReturn